/*
 * Copyright 2020 University of Liège
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include "wF0El.h"
using namespace tbox;
using namespace dart;

/**
 * @brief Update the value
 */
void F0ElC::update(double _v)
{
    v = _v;
}
/**
 * @brief Evaluate the constant
 */
double F0ElC::eval(Element const &e, std::vector<double> const &phi, size_t k) const
{
    return v;
}
/**
 * @brief Evaluate the gradient of the constant
 */
double F0ElC::evalGrad(Element const &e, std::vector<double> const &phi, size_t k) const
{
    return 0;
}

/**
 * @brief Precompute values for Padé approximation
 *
 * The input _mC2 is the square of the critical Mach number (used to compute
 * the critical velocity).
 */
F0ElRhoBase::F0ElRhoBase(double _mInf, double _mC2) : F0El(), gamma(1.4), mInf(_mInf)
{
    uC = sqrt(_mC2 * (1 / (mInf * mInf) + (gamma - 1) / 2) / (1 + (gamma - 1) / 2 * _mC2)); // critical velocity
    rC = 1 + 0.5 * (gamma - 1) * mInf * mInf * (1 - uC * uC);                               // r(uC)
    beta = ((gamma - 1) * mInf * mInf * uC) / rC;                                           // dr/du(uC) / -r(uC)
}
/**
 * @brief Evaluate the base of the isentropic density
 *
 * r = [1 + (gamma-1)/2*Mi^2*(1-u^2)], u <= uC
 * r = r(uC) / (1-(dr/du(uC)/f(uC))*(u/uC-1)), u > uC
 */
double F0ElRhoBase::eval(Element const &e, std::vector<double> const &phi, size_t k) const
{
    double u = e.computeGradient(phi, k).norm(); // norm of velocity
    if (u <= uC)
        return 1 + 0.5 * (gamma - 1) * mInf * mInf * (1 - u * u);
    else
        return rC / (1 + beta * (u / uC - 1));
}
/**
 * @brief Evaluate the gradient of the base of the isentropic density
 *
 * dr = -(gamma-1)*Mi^2 * u, u <= uC
 * dr = -r(uC) / (1-(dr/du(uC)/f(uC))*(u/uC-1))^2 * (dr/du(uC)/f(uC)) / uC, u > uC
 * Computed value has to be multiplied by u to recover actual gradient
 */
double F0ElRhoBase::evalGrad(Element const &e, std::vector<double> const &phi, size_t k) const
{
    double u = e.computeGradient(phi, k).norm(); // norm of velocity
    if (u <= uC)
        return -(gamma - 1) * mInf * mInf;
    else
        return -rC / ((1 + beta * (u / uC - 1)) * (1 + beta * (u / uC - 1))) * beta / uC / u;
}

/**
 * @brief Evaluate the isentropic density
 *
 * rho = (1 + (gamma-1)/2*Mi^2*(1-u^2))^(1/(gamma-1))
 */
double F0ElRho::eval(Element const &e, std::vector<double> const &phi, size_t k) const
{
    double u = e.computeGradient(phi, k).norm(); // norm of velocity
    double r = 1 + 0.5 * (gamma - 1) * mInf * mInf * (1 - u * u);
    return sqrt(r * r * r * r * r); // particularized to gamma=1.4
}
/**
 * @brief Evaluate the isentropic density gradient
 *
 * drho = -Mi^2 * rho^(2-gamma) * u
 * Computed value has to be multiplied by u to recover actual gradient
 */
double F0ElRho::evalGrad(Element const &e, std::vector<double> const &phi, size_t k) const
{
    double u = e.computeGradient(phi, k).norm(); // norm of velocity
    double r = 1 + 0.5 * (gamma - 1) * mInf * mInf * (1 - u * u);
    return -mInf * mInf * sqrt(r * r * r); // particularized to gamma=1.4
}

/**
 * @brief Evaluate the (clamped) isentropic density
 *
 * rho = r^(1/(gamma-1))
 */
double F0ElRhoClamp::eval(Element const &e, std::vector<double> const &phi, size_t k) const
{
    double r = F0ElRhoBase::eval(e, phi, k);
    return sqrt(r * r * r * r * r); // particularized to gamma=1.4
}
/**
 * @brief Evaluate the (clamped) isentropic density gradient
 *
 * drho = 1/(gamma-1) * r^(2-gamma) * dr
 * Computed value has to be multiplied by u to recover actual gradient
 */
double F0ElRhoClamp::evalGrad(Element const &e, std::vector<double> const &phi, size_t k) const
{
    double r = F0ElRhoBase::eval(e, phi, k);
    double dr = F0ElRhoBase::evalGrad(e, phi, k);
    return 1 / (gamma - 1) * sqrt(r * r * r) * dr; // particularized to gamma=1.4
}

/**
 * @brief Evaluate the linear density
 */
double F0ElRhoLin::eval(Element const &e, std::vector<double> const &phi, size_t k) const
{
    return 1.;
}
/**
 * @brief Evaluate the linear density gradient
 */
double F0ElRhoLin::evalGrad(Element const &e, std::vector<double> const &phi, size_t k) const
{
    return 0.;
}

/**
 * @brief Evaluate the isentropic Mach number
 *
 * M = u / sqrt(1/mi^2 + (gamma-1)/2*(1-u^2))
 */
double F0ElMach::eval(Element const &e, std::vector<double> const &phi, size_t k) const
{
    double u2 = e.computeGradient(phi, k).squaredNorm();          // velocity squared
    double a2 = 1 / (mInf * mInf) + 0.5 * (gamma - 1) * (1 - u2); // speed of sound squared
    return sqrt(u2) / sqrt(a2);
}
/**
 * @brief Evaluate the isentropic Mach number gradient
 *
 * dM = (1/(u*a) + (gamma-1)/2*u/a^3) * u
 * Computed value has to be multiplied by u to recover actual gradient
 */
double F0ElMach::evalGrad(Element const &e, std::vector<double> const &phi, size_t k) const
{
    double u2 = e.computeGradient(phi, k).squaredNorm();          // velocity squared
    double a2 = 1 / (mInf * mInf) + 0.5 * (gamma - 1) * (1 - u2); // speed of sound squared
    return 1 / sqrt(u2 * a2) + 0.5 * (gamma - 1) * sqrt(u2) / sqrt(a2 * a2 * a2);
}

/**
 * @brief Evaluate the (clamped) isentropic Mach number
 *
 * M = u / sqrt(r/Mi^2)
 */
double F0ElMachClamp::eval(Element const &e, std::vector<double> const &phi, size_t k) const
{
    double u = e.computeGradient(phi, k).norm(); // norm of velocity
    double r = F0ElRhoBase::eval(e, phi, k);
    return mInf * u / sqrt(r);
}
/**
 * @brief Evaluate the (clamped) isentropic Mach number gradient
 *
 * dM = 1 / sqrt(r/Mi^2) + u / sqrt(r^3/Mi^2) * dr
 * Computed value has to be multiplied by u to recover actual gradient
 */
double F0ElMachClamp::evalGrad(Element const &e, std::vector<double> const &phi, size_t k) const
{
    double u = e.computeGradient(phi, k).norm(); // norm of velocity
    double r = F0ElRhoBase::eval(e, phi, k);
    double dr = F0ElRhoBase::evalGrad(e, phi, k);
    return mInf * (1 / sqrt(r) / u - 0.5 * u / sqrt(r * r * r) * dr);
}

/**
 * @brief Evaluate the linear Mach number
 */
double F0ElMachLin::eval(Element const &e, std::vector<double> const &phi, size_t k) const
{
    return 0.;
}
/**
 * @brief Evaluate the linear Mach number gradient
 */
double F0ElMachLin::evalGrad(Element const &e, std::vector<double> const &phi, size_t k) const
{
    return 0.;
}

/**
 * @brief Evaluate the isentropic pressure coefficient
 *
 * Cp = 2/(gamma*Mi^2) * ((1 + (gamma-1)/2*Mi^2*(1-u^2))^(gamma/(gamma-1)) - 1)
 */
double F0ElCp::eval(Element const &e, std::vector<double> const &phi, size_t k) const
{
    double u2 = e.computeGradient(phi, k).squaredNorm(); // velocity squared
    double r = 1 + 0.5 * (gamma - 1) * mInf * mInf * (1 - u2);
    return 2 / (gamma * mInf * mInf) * (sqrt(r * r * r * r * r * r * r) - 1); // particularized to gamma=1.4
}
/**
 * @brief Evaluate the isentropic pressure coefficient gradient
 *
 * dCp = -2*(1 + (gamma-1)/2*Mi^2*(1-u^2))^(1/(gamma-1)) * u
 * Computed value has to multiplied by u to recover actual gradient
 */
double F0ElCp::evalGrad(Element const &e, std::vector<double> const &phi, size_t k) const
{
    double u2 = e.computeGradient(phi, k).squaredNorm(); // velocity squared
    double r = 1 + 0.5 * (gamma - 1) * mInf * mInf * (1 - u2);
    return -2 * sqrt(r * r * r * r * r); // particularized to gamma=1.4
}

/**
 * @brief Evaluate the (clamped) isentropic pressure coefficient
 *
 * Cp = 2/(gamma*Mi^2) * (r^(gamma/(gamma-1)) - 1)
 */
double F0ElCpClamp::eval(Element const &e, std::vector<double> const &phi, size_t k) const
{
    double r = F0ElRhoBase::eval(e, phi, k);
    return 2 / (gamma * mInf * mInf) * (sqrt(r * r * r * r * r * r * r) - 1); // particularized to gamma=1.4
}
/**
 * @brief Evaluate the (clamped) isentropic pressure coefficient gradient
 *
 * dCp = 2/((gamma-1)*Mi^2) * r^(1/(gamma-1)) * dr
 * Computed value has to multiplied by u to recover actual gradient
 */
double F0ElCpClamp::evalGrad(Element const &e, std::vector<double> const &phi, size_t k) const
{
    double r = F0ElRhoBase::eval(e, phi, k);
    double dr = F0ElRhoBase::evalGrad(e, phi, k);
    return 2 / ((gamma - 1) * mInf * mInf) * sqrt(r * r * r * r * r) * dr; // particularized to gamma=1.4
}

/**
 * @brief Evaluate the linear pressure coefficient
 *
 * Cp = 1-u^2
 */
double F0ElCpLin::eval(Element const &e, std::vector<double> const &phi, size_t k) const
{
    return 1 - e.computeGradient(phi, k).squaredNorm();
}
/**
 * @brief Evaluate the linear pressure coefficient gradient
 *
 * dCp = -2*u
 * Computed value has to multiplied by u to recover actual gradient
 */
double F0ElCpLin::evalGrad(Element const &e, std::vector<double> const &phi, size_t k) const
{
    return -2.;
}