/*
 * Copyright 2020 University of Liège
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#ifndef WF1CT_H
#define WF1CT_H

#include "dart.h"
#include "wObserver.h"
#include <Eigen/Dense>

namespace dart
{

/**
 * @brief Constant vector function
 */
class DART_API F1Ct : public fwk::Observer
{
public:
    F1Ct() : Observer() {}
    virtual ~F1Ct() {}

    virtual Eigen::Vector3d eval() const = 0;
    virtual Eigen::Vector3d evalGrad() const = 0;
};

/**
 * @brief Drag direction
 */
class DART_API F1CtDrag : public F1Ct
{
    int nDim;           ///< number of dimensions
    double sRef;        ///< reference area
    double beta;        ///< angle of sideslip
    Eigen::Vector3d v;  ///< drag direction unit vector
    Eigen::Vector3d dv; ///< gradient of drag direction unit vector (wrt alpha)

public:
    F1CtDrag(int _nDim, double _sRef, double alpha, double _beta = 0);
    virtual void update(double alpha) override;

    virtual Eigen::Vector3d eval() const override;
    virtual Eigen::Vector3d evalGrad() const override;
};

/**
 * @brief Sideforce direction
 */
class DART_API F1CtSide : public F1Ct
{
    int nDim;           ///< number of dimensions
    double sRef;        ///< reference area
    double beta;        ///< angle of sideslip
    Eigen::Vector3d v;  ///< sideforce direction unit vector
    Eigen::Vector3d dv; ///< gradient of sideforce direction unit vector (wrt alpha)

public:
    F1CtSide(int _nDim, double _sRef, double alpha, double _beta = 0);
    virtual void update(double alpha) override;

    virtual Eigen::Vector3d eval() const override;
    virtual Eigen::Vector3d evalGrad() const override;
};

/**
 * @brief Lift direction
 */
class DART_API F1CtLift : public F1Ct
{
    int nDim;           ///< number of dimensions
    double sRef;        ///< reference area
    double beta;        ///< angle of sideslip
    Eigen::Vector3d v;  ///< lift direction unit vector
    Eigen::Vector3d dv; ///< gradient of lift direction unit vector (wrt alpha)

public:
    F1CtLift(int _nDim, double _sRef, double alpha, double _beta = 0);
    virtual void update(double alpha) override;

    virtual Eigen::Vector3d eval() const override;
    virtual Eigen::Vector3d evalGrad() const override;
};

} // namespace dart

#endif //WF1CT_H
