################################################################################
#                                                                              #
#                            DARTFlo viscous module file                       #
#                          Variables: Solution and external                    #
#                             flow parameters container                        #
#                                                                              #
# Author: Paul Dechamps                                                        #
# Université de Liège                                                          #
# Date: 2021.09.10                                                             #
#                                                                              #
################################################################################


# ------------------------------------------------------------------------------
#  Imports
# ------------------------------------------------------------------------------
import numpy as np

class PGeometry:

    def __init__(self, data, paramDict, _bNodes):

        # Private
        
        self.__globCoord     = data[:,0]                        # Global coordinates of each marker
        self.__locCoord      = paramDict['xx']                  # Local coordinates of each marker
        self.__nMarkers      = len(self.__locCoord)             # Number of markers in the computational domain
        
        self.__boundaryNodes = _bNodes                          # Boundary nodes of the computational domain [Stagnation point,
                                                                #                                             First point on lower side,
                                                                #                                             First wake point]

    def GetglobCoord(self): return self.__globCoord
    def GetlocCoord(self): return self.__locCoord
    def GetnMarkers(self): return self.__nMarkers
    def GetglobCoord(self): return self.__bNodes
    def GetboundaryNodes(self): return self.__boundaryNodes


class PSolution:

    def __init__(self, _xtrF):

        # Private

        self.__nVar = 5                                           # Number of variables involved in the system of PDEs
        self.__nBlPar = 10                                        # Number of boundary layer parameters
        
        self.__xtrF = _xtrF                                       # Forced Transition location on [upper, lower] sides
        self.__xtr = [1, 1]                                       # Transition location

        # Public

        # Solution vectors

        self.u = []                                               # Vector containing the solution [theta, H, N, ue, Ctau]

        # Boundary layer parameters

        self.ReTheta =                                            # Reynolds number based on the variable 'theta'.
        self.disspCoeff =                                         # Dissipation coefficient.
        self.frictionCoeff =                                      # Wall friction coefficient.
        self.delta =                                              # Boundary layer thickness.
        self.Hstar =                                              # Kinetic energy shape parameter.
        self.Hstar2 =                                             # Density shape parameter.
        self.Hk =                                                 # Kinematic shape factor.
        self.Cteq =                                               # Equilibrium shear stress coefficient.
        self.Us =                                                 # Equivalent normalized wall slip velocity.
        self.deltaStar =                                          # Displacement thickness of the boundary layer. 

    def GetnVar(self): return self.__nVar
    def GetnBlPar(self): return self.__nBlPar

    def GetxtrF(self): return self.__xtrF
    def Getxtr(self): return self.__xtr
    def Setxtr(self, xtrIn, side):
        self.__xtr[side] = xtrIn
        pass

    def initializeSolution(self, PGeometry):

        nMarkers = PGeometry.GetnMarkers()
        self.u = np.zeros(self.__nVar * nMarkers)
        self.blPar = np.zeros(self.__nVar * nMarkers)

        return self.u, self.blPar

class POuterFlow:

    def __init__(self, data) -> None:

        self.__Mach =                                               # Outer Mach number.
        self.__density =                                            # Outer flow density.
        self.__velocity =                                           # Outer flow velocity.
        self.__dv =                                                 # Outer flow velocity gradient. 

        self.outerDeltaStar = 
        self.outerLocCoord = 
        self.outerRe = 

class PDataContainer:
    def __init__(self, _Re) -> None:

        self.Re         = _Re               # Flow reynolds number.
        self.Geometry  = PGeometry()       # Geometry components.
        self.Solution  = PSolution()       # Solution components.
        self.OuterFlow = POuterFlow()      # Outer (inviscid) flow parameters.


        


class DVariables:
    """Data container for viscous solver.

    Attributes
    ----------
    
    - Ti: float [0,1]
        - User defined turbulence intensity that impacts the critical amplification ratio Ncrit of the e^N method.
    - Re: float
        - Flow Reynolds number.

    - xx: numpy array
        - Airfoil surfaces and wake mesh.

    - nN: int
        - Number of cell in the computational domain.

    - bNodes: numpy array
            - 3 elements array containing the boundary nodes of the computational domain [Stagnation point, First lower side point, First wake point].
    
    - nVar: int
          - Number of variables involved in the differential problem.

    - u: numpy array (vector of len(nVar*nN))
         - Array containing the solution (nVar variables at each of the nN points).

    - blPar: numpy array (vector of len(nBlPar*nN))
           - Boundary layer parameters.

    - extPar: numpy array (vector of len(nExtPar*nN))
            - External (inviscid) flow parameters.

    - dv: numpy array (vector of len(nN))
        - Velocity gradient on each cell.

    - xGlobal: numpy array (vector of len(nN))
             - Cell position non-dimensionalized by the airfoil chord: x/c

    - xtrF: numpy array of len(2)
          - User forced transiton on [Upper side, Lower side]
    
    - xtr: numpy array of len(2)
         - Flow laminar to turbulent transition location on [Upper side, Lower side]
        
    - Ncrit: float
            - Critical amplification ratio for the e^N method (default value= 9.) 

    """
    def __init__(self, data, paramDict, bNodes, xtrF, Ti, Re):
        self.__Ti=Ti
        self.Re=Re
        self.xx=paramDict['xx']
        self.nN=len(self.xx)

        self.bNodes=bNodes
        # Solution 
        self.nVar=5
        self.u=np.zeros(self.nVar*self.nN)

        # Parameters of the boundary layers
        # [ 0 ,  1,  2,     3,     4,      5,  6,    7,  8,         9]
        # [Ret, Cd, Cf, delta, Hstar, Hstar2, Hk, Cteq, Us, deltaStar]
        self.nBlPar=10
        self.blPar=np.zeros(self.nBlPar*self.nN)
        self.Ret_previous=self.blPar[0::self.nVar]

        self.dueOld_dx=np.zeros(self.nN)
        self.ddeltaStarOld_dx=np.zeros(self.nN)

        # Parameters of the external (inviscid) flow
        # [ 0,    1,     2,            3,     4,    5]
        # [Me, rhoe, vtExt, deltaStarExt, xxExt ReExt]
        self.nExtPar=6
        self.extPar=np.zeros(self.nExtPar*self.nN)
        self.extPar[0::self.nExtPar]=data[:,5]
        self.extPar[1::self.nExtPar]=data[:,6]
        self.extPar[2::self.nExtPar]=paramDict['vtExt']
        self.extPar[3::self.nExtPar]=data[:,7]
        self.extPar[4::self.nExtPar]=paramDict['xx']
        self.dv=paramDict['dv']
        self.xGlobal=data[:,0]

        self.turb=np.zeros(self.nN, dtype=int)

        self.xtrF=xtrF
        self.xtr=np.zeros(2)
        if not len(self.xtrF)==2:
            raise RuntimeError(ccolors.ANSI_RED + "Error in forced transition." + ccolors.ANSI_RESET)
        for n in range(len(self.xtrF)):
            if self.xtrF[n] is not None:
                if 0<=self.xtrF[n]<=1:
                    self.xtr[n]=self.xtrF[n]
                else:
                    raise ValueError(ccolors.ANSI_RED + "Non-physical transition location." + ccolors.ANSI_RESET)
            else:
                self.xtr[n]=1

        self.idxTr=[self.bNodes[1]-1,self.bNodes[2]-1]

        self.Ncrit=self.turbIntensity(Ti)
        pass
    
    def turbIntensity(self,Ti):
        '''Computes the critical amplification ratio Ncrit based on the input
        free-stream turbulence intensity. Ncrit=9 by default.'''
        if Ti is not None and Ti<=1:
            tau=2.7*np.tanh(self.Ti/2.7)
            Ncrit=-8.43-2.4*np.log(tau/100) # Drela 1998
        else:
            Ncrit=9
        return Ncrit