#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright 2020 University of Liège
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


## VII Coupler
# Paul Dechamps

import dart.utils as floU
import tbox.utils as tboxU

from vii.pyVII.DAirfoil import Airfoil
from vii.pyVII.DWake import Wake
from vii.pyVII.DGroupConstructor import GroupConstructor

from fwk.coloring import ccolors
import math

import fwk
#import dart.viscUtils as viscU
import numpy as np
from matplotlib import pyplot as plt

class Coupler:
  def __init__(self, iSolver, vSolver, _iterPrint=1) -> None:
    self.iSolver=iSolver
    self.vSolver=vSolver

    self.maxCouplIter = 150
    self.couplTol = 1e-4
    self.iterPrint = _iterPrint

    self.tms=fwk.Timers()
    self.resetInv = True
    pass

  def CreateProblem(self, _boundaryAirfoil, _boundaryWake):

    self.group = [Airfoil(_boundaryAirfoil), Wake(_boundaryWake)] # airfoil and wake python objects
    self.blwVel = [None, None, None]
    self.deltaStar = [None, None, None]
    self.xx = [None, None, None]
    pass

  def Run(self):

    # Initilize meshes in c++ objects

    aeroCoeffs = np.zeros((0,2))

    self.nElmUpperPrev = 0
    couplIter = 0
    cdPrev = 0.0

    print('- AoA: {:<2.2f} Mach: {:<1.1f} Re: {:<2.1f}e6'
    .format(self.iSolver.pbl.alpha*180/math.pi, self.iSolver.pbl.M_inf, self.vSolver.GetRe()/1e6))
    print('{:>5s}| {:>10s} {:>10s} | {:>8s} {:>6s} {:>8s}'.format('It', 'Cl', 'Cd', 'xtrT', 'xtrB', 'Error'))
    print(' --------------------------------------------------------')

    while couplIter < self.maxCouplIter:

      # Run inviscid solver
      self.tms['inviscid'].start()
      if self.resetInv:
        self.iSolver.reset()
      self.iSolver.run()
      self.tms['inviscid'].stop()

      if couplIter == 0:
        # Write inviscid Cp distribution
        Cp = floU.extract(self.group[0].boundary.tag.elems, self.iSolver.Cp)
        tboxU.write(Cp, 'Cp_Inviscid.dat', '%1.5e', ', ', 'x, y, z, Cp', '')

      # Extract inviscid boundary
      self.tms['processing'].start()
      self.__ImposeInvBC(couplIter)
      self.tms['processing'].stop()

      # Run viscous solver
      self.tms['viscous'].start()
      exitCode = self.vSolver.Run(couplIter)
      self.tms['viscous'].stop()

      self.tms['processing'].start()
      self.__ImposeBlwVel(couplIter)
      self.tms['processing'].stop()

      error = abs((self.vSolver.Cdt - cdPrev)/self.vSolver.Cdt) if self.vSolver.Cdt != 0 else 1

      aeroCoeffs = np.vstack((aeroCoeffs, [self.iSolver.Cl, self.vSolver.Cdt]))
      
      if error  <= self.couplTol:
        print('{:>5s}| {:>10s} {:>10s} | {:>10s} {:>8s} {:>8s}'.format('It', 'Cl', 'Cd', 'xtrT', 'xtrB', 'Error'))
        print(ccolors.ANSI_GREEN,'{:>4.0f}| {:>10.5f} {:>10.5f} | {:>10.4f} {:>8.4f} {:>8.2f}\n'.format(couplIter,
        self.iSolver.Cl, self.vSolver.Cdt, self.vSolver.Getxtr(0,0), self.vSolver.Getxtr(0,1), np.log10(error)),ccolors.ANSI_RESET)
        return aeroCoeffs
      
      cdPrev = self.vSolver.Cdt

      if couplIter%self.iterPrint==0:
        print(' {:>4.0f}| {:>10.4f} {:>10.4f} | {:>10.4f} {:>8.4f} {:>8.3f}'.format(couplIter,
        self.iSolver.Cl, self.vSolver.Cdt, self.vSolver.Getxtr(0,0), self.vSolver.Getxtr(0,1), np.log10(error)))

        if self.iSolver.verbose != 0 or self.vSolver.verbose != 0:
          print('')

      couplIter += 1
      self.tms['processing'].stop()
      
    else:
      print(ccolors.ANSI_RED, 'Maximum number of {:<.0f} coupling iterations reached'.format(self.maxCouplIter), ccolors.ANSI_RESET)
      return aeroCoeffs


  def RunPolar(self, alphaSeq):
    
    import math

    self.iSolver.printOn = False
    self.vSolver.printOn = False

    coeffTable = np.zeros((3,len(alphaSeq)))
    coeffTable[0,:] = alphaSeq

    for ind, alpha in enumerate(alphaSeq):
      self.iSolver.pbl.update(alpha*math.pi/180)

      self.Run()

      # Collect Coeff

      coeffTable[1,ind] = self.iSolver.Cl
      coeffTable[2,ind] = self.vSolver.Cdt

    return coeffTable

  def __ImposeBlwVel(self, couplIter):

    # Retreive and set blowing velocities.
    for iRegion in range(3):
      self.blwVel[iRegion] = np.asarray(self.vSolver.ExtractBlowingVel(0, iRegion))
      self.deltaStar[iRegion] = np.asarray(self.vSolver.ExtractDeltaStar(0, iRegion))
      self.xx[iRegion] = np.asarray(self.vSolver.Extractxx(0, iRegion))

    blwVelAF = np.concatenate((self.blwVel[0], self.blwVel[1]))
    deltaStarAF = np.concatenate((self.deltaStar[0], self.deltaStar[1][1:])) # Remove stagnation point doublon.
    xxAF = np.concatenate((self.xx[0], self.xx[1][1:])) # Remove stagnation point doublon.
    blwVelWK = self.blwVel[2]
    deltaStarWK = self.deltaStar[2]
    xxWK = self.xx[2]

    self.group[0].u = blwVelAF[self.group[0].connectListElems.argsort()]
    self.group[1].u = blwVelWK[self.group[1].connectListElems.argsort()]

    self.group[0].deltaStar = deltaStarAF[self.group[0].connectListNodes.argsort()]
    self.group[1].deltaStar = deltaStarWK[self.group[1].connectListNodes.argsort()]

    self.group[0].xx = xxAF[self.group[0].connectListNodes.argsort()]
    self.group[1].xx = xxWK[self.group[1].connectListNodes.argsort()]

    for n in range(0, len(self.group)):
      for i in range(0, self.group[n].nE):
        self.group[n].boundary.setU(i, self.group[n].u[i])

  def __ImposeInvBC(self, couplIter):

    for n in range(0, len(self.group)):

      for i in range(0, len(self.group[n].boundary.nodes)):

        self.group[n].v[i,0] = self.iSolver.U[self.group[n].boundary.nodes[i].row][0]
        self.group[n].v[i,1] = self.iSolver.U[self.group[n].boundary.nodes[i].row][1]
        self.group[n].v[i,2] = self.iSolver.U[self.group[n].boundary.nodes[i].row][2]
        self.group[n].Me[i] = self.iSolver.M[self.group[n].boundary.nodes[i].row]
        self.group[n].rhoe[i] = self.iSolver.rho[self.group[n].boundary.nodes[i].row]
    
    dataIn=[None,None]
    for n in range(0,len(self.group)):
      self.group[n].connectListNodes, self.group[n].connectListElems, dataIn[n]  = self.group[n].connectList(couplIter)
    self.fMeshDict, _, _ = GroupConstructor().mergeVectors(dataIn, 0, 1)
    fMeshDict = self.fMeshDict.copy()
    if ((len(fMeshDict['locCoord'][:fMeshDict['bNodes'][1]]) != self.nElmUpperPrev) or couplIter==0):
      
      # Initialize mesh

      self.vSolver.SetGlobMesh(0, 0, fMeshDict['globCoord'][:fMeshDict['bNodes'][1]], fMeshDict['yGlobCoord'][:fMeshDict['bNodes'][1]], fMeshDict['zGlobCoord'][:fMeshDict['bNodes'][1]])
      self.vSolver.SetGlobMesh(0, 1, fMeshDict['globCoord'][fMeshDict['bNodes'][1]:fMeshDict['bNodes'][2]], fMeshDict['yGlobCoord'][fMeshDict['bNodes'][1]:fMeshDict['bNodes'][2]], fMeshDict['zGlobCoord'][fMeshDict['bNodes'][1]:fMeshDict['bNodes'][2]])
      self.vSolver.SetGlobMesh(0, 2, fMeshDict['globCoord'][fMeshDict['bNodes'][2]:], fMeshDict['yGlobCoord'][fMeshDict['bNodes'][2]:], fMeshDict['zGlobCoord'][fMeshDict['bNodes'][2]:])
      #self.vSolver.InitMeshes(0, fMeshDict['locCoord'][:fMeshDict['bNodes'][1]], fMeshDict['globCoord'][:fMeshDict['bNodes'][1]])
      #self.vSolver.InitMeshes(1, fMeshDict['locCoord'][fMeshDict['bNodes'][1]:fMeshDict['bNodes'][2]], fMeshDict['globCoord'][fMeshDict['bNodes'][1]:fMeshDict['bNodes'][2]])
      #self.vSolver.InitMeshes(2, fMeshDict['locCoord'][fMeshDict['bNodes'][2]:], fMeshDict['globCoord'][fMeshDict['bNodes'][2]:])

      self.vSolver.SetxxExt(0, 0, fMeshDict['locCoord'][:fMeshDict['bNodes'][1]])
      self.vSolver.SetxxExt(0, 1, fMeshDict['locCoord'][fMeshDict['bNodes'][1]:fMeshDict['bNodes'][2]])    
      self.vSolver.SetxxExt(0, 2, fMeshDict['locCoord'][fMeshDict['bNodes'][2]:])
      
      self.vSolver.SetDeltaStarExt(0, 0, fMeshDict['deltaStarExt'][:fMeshDict['bNodes'][1]])
      self.vSolver.SetDeltaStarExt(0, 1, fMeshDict['deltaStarExt'][fMeshDict['bNodes'][1]:fMeshDict['bNodes'][2]])    
      self.vSolver.SetDeltaStarExt(0, 2, fMeshDict['deltaStarExt'][fMeshDict['bNodes'][2]:])

    else:

      self.vSolver.SetxxExt(0, 0, fMeshDict['xxExt'][:fMeshDict['bNodes'][1]])
      self.vSolver.SetxxExt(0, 1, fMeshDict['xxExt'][fMeshDict['bNodes'][1]:fMeshDict['bNodes'][2]])    
      self.vSolver.SetxxExt(0, 2, fMeshDict['xxExt'][fMeshDict['bNodes'][2]:])
      
      self.vSolver.SetDeltaStarExt(0, 0, fMeshDict['deltaStarExt'][:fMeshDict['bNodes'][1]])
      self.vSolver.SetDeltaStarExt(0, 1, fMeshDict['deltaStarExt'][fMeshDict['bNodes'][1]:fMeshDict['bNodes'][2]])    
      self.vSolver.SetDeltaStarExt(0, 2, fMeshDict['deltaStarExt'][fMeshDict['bNodes'][2]:])

    """plt.plot(fMeshDict['globCoord'][:fMeshDict['bNodes'][1]], fMeshDict['vx'][:fMeshDict['bNodes'][1]], 'x')
    plt.plot(fMeshDict['globCoord'][fMeshDict['bNodes'][1]:fMeshDict['bNodes'][2]], fMeshDict['vx'][fMeshDict['bNodes'][1]:fMeshDict['bNodes'][2]], 'x')
    plt.plot(fMeshDict['globCoord'][fMeshDict['bNodes'][2]:], fMeshDict['vx'][fMeshDict['bNodes'][2]:], 'x')
    plt.title('Ux = f(x)')
    plt.show()
    plt.plot(fMeshDict['globCoord'][:fMeshDict['bNodes'][1]], fMeshDict['vy'][:fMeshDict['bNodes'][1]], 'x')
    plt.plot(fMeshDict['globCoord'][fMeshDict['bNodes'][1]:fMeshDict['bNodes'][2]], fMeshDict['vy'][fMeshDict['bNodes'][1]:fMeshDict['bNodes'][2]], 'x')
    plt.plot(fMeshDict['globCoord'][fMeshDict['bNodes'][2]:], fMeshDict['vy'][fMeshDict['bNodes'][2]:], 'x')
    plt.title('Uy = f(x)')
    plt.show()
    plt.plot(fMeshDict['globCoord'][:fMeshDict['bNodes'][1]], fMeshDict['vz'][:fMeshDict['bNodes'][1]], 'x')
    plt.plot(fMeshDict['globCoord'][fMeshDict['bNodes'][1]:fMeshDict['bNodes'][2]], fMeshDict['vz'][fMeshDict['bNodes'][1]:fMeshDict['bNodes'][2]], 'x')
    plt.plot(fMeshDict['globCoord'][fMeshDict['bNodes'][2]:], fMeshDict['vz'][fMeshDict['bNodes'][2]:], 'x')
    plt.title('Uz = f(x)')
    plt.show()"""

    """plt.plot(fMeshDict['globCoord'][:fMeshDict['bNodes'][1]], fMeshDict['vy'][:fMeshDict['bNodes'][1]], 'x-')
    plt.plot(fMeshDict['globCoord'][fMeshDict['bNodes'][1]:fMeshDict['bNodes'][2]], fMeshDict['vy'][fMeshDict['bNodes'][1]:fMeshDict['bNodes'][2]], 'x-')
    #plt.plot(fMeshDict['globCoord'][fMeshDict['bNodes'][2]:], fMeshDict['vy'][fMeshDict['bNodes'][2]:], 'x-')
    plt.xlim([0.4, 1.2])
    plt.ylim([-0.3, 0.2])
    plt.title('Uy = f(x)')
    plt.show()"""
    """plt.plot(fMeshDict['globCoord'][:fMeshDict['bNodes'][1]], fMeshDict['vx'][:fMeshDict['bNodes'][1]])
    plt.plot(fMeshDict['globCoord'][fMeshDict['bNodes'][1]:fMeshDict['bNodes'][2]], fMeshDict['vx'][fMeshDict['bNodes'][1]:fMeshDict['bNodes'][2]])
    plt.show()"""
    
    self.nElmUpperPrev = len(fMeshDict['locCoord'][:fMeshDict['bNodes'][1]])
    self.vSolver.SetVelocities(0, 0, fMeshDict['vx'][:fMeshDict['bNodes'][1]], fMeshDict['vy'][:fMeshDict['bNodes'][1]], fMeshDict['vz'][:fMeshDict['bNodes'][1]])
    self.vSolver.SetVelocities(0, 1, fMeshDict['vx'][fMeshDict['bNodes'][1]:fMeshDict['bNodes'][2]], fMeshDict['vy'][fMeshDict['bNodes'][1]:fMeshDict['bNodes'][2]], fMeshDict['vz'][fMeshDict['bNodes'][1]:fMeshDict['bNodes'][2]])
    self.vSolver.SetVelocities(0, 2, fMeshDict['vx'][fMeshDict['bNodes'][2]:], fMeshDict['vy'][fMeshDict['bNodes'][2]:], fMeshDict['vz'][fMeshDict['bNodes'][2]:])
    #self.vSolver.SendInvBC(0, fMeshDict['vtExt'][:fMeshDict['bNodes'][1]], fMeshDict['Me'][:fMeshDict['bNodes'][1]], fMeshDict['rho'][:fMeshDict['bNodes'][1]])
    #self.vSolver.SendInvBC(1, fMeshDict['vtExt'][fMeshDict['bNodes'][1]:fMeshDict['bNodes'][2]], fMeshDict['Me'][fMeshDict['bNodes'][1]:fMeshDict['bNodes'][2]], fMeshDict['rho'][fMeshDict['bNodes'][1]:fMeshDict['bNodes'][2]])
    #self.vSolver.SendInvBC(2, fMeshDict['vtExt'][fMeshDict['bNodes'][2]:], fMeshDict['Me'][fMeshDict['bNodes'][2]:], fMeshDict['rho'][fMeshDict['bNodes'][2]:])
    self.vSolver.SetMe(0, 0, fMeshDict['Me'][:fMeshDict['bNodes'][1]])
    self.vSolver.SetMe(0, 1, fMeshDict['Me'][fMeshDict['bNodes'][1]:fMeshDict['bNodes'][2]])
    self.vSolver.SetMe(0, 2, fMeshDict['Me'][fMeshDict['bNodes'][2]:])
    self.vSolver.SetRhoe(0, 0, fMeshDict['rho'][:fMeshDict['bNodes'][1]])
    self.vSolver.SetRhoe(0, 1, fMeshDict['rho'][fMeshDict['bNodes'][1]:fMeshDict['bNodes'][2]])
    self.vSolver.SetRhoe(0, 2, fMeshDict['rho'][fMeshDict['bNodes'][2]:])
    # print('Inviscid boundary imposed')
  


  



