#define _USE_MATH_DEFINES

#include "wViscFluxes.h"
#include "wBLRegion.h"
#include <Eigen/Dense>

using namespace vii;
using namespace Eigen;

ViscFluxes::ViscFluxes(double _Re)
{
    Re = _Re;

}

ViscFluxes::~ViscFluxes()
{
    std::cout << "~ViscFluxes()\n";
}

VectorXd ViscFluxes::ComputeFluxes(size_t iPoint, BLRegion *bl)
{
    unsigned int nVar = bl->GetnVar();
    std::vector<double> u(nVar, 0.);
    for (size_t k = 0; k < nVar; ++k)
        u[k] = bl->U[iPoint * nVar + k];
    return BLlaws(iPoint, bl, u);
}

MatrixXd ViscFluxes::ComputeJacobian(size_t iPoint, BLRegion *bl, VectorXd &sysRes, double eps)
{

    unsigned int nVar = bl->GetnVar();
    MatrixXd JacMatrix = MatrixXd::Zero(5,5);
    std::vector<double> uUp(nVar, 0.);
    for (size_t k = 0; k < nVar; ++k)
        uUp[k] = bl->U[iPoint * nVar + k];

    double varSave = 0.;
    for (size_t iVar = 0; iVar < nVar; ++iVar)
    {
        varSave = uUp[iVar];
        uUp[iVar] += eps;
        JacMatrix.col(iVar) = (BLlaws(iPoint, bl, uUp) - sysRes) / eps;
        uUp[iVar] = varSave;
    }
    return JacMatrix;
}

VectorXd ViscFluxes::BLlaws(size_t iPoint, BLRegion *bl, std::vector<double> u)
{
    unsigned int nVar = bl->GetnVar();

    MatrixXd timeMatrix = MatrixXd::Zero(5, 5);
    VectorXd spaceVector = VectorXd::Zero(5);

    double dissipRatio;
    if (bl->name == "wake")
        dissipRatio = 0.9;
    else
        dissipRatio = 1.;

    bl->Ret[iPoint] = std::max(u[3] * u[0] * Re, 1.0);  /* Reynolds number based on theta */
    bl->DeltaStar[iPoint] = u[0] * u[1];                /* Displacement thickness */

    // Boundary layer closure relations.
    if (bl->Regime[iPoint] == 0)
    {
        // Laminar closure relations.
        std::vector<double> lamParam(8, 0.);
        bl->closSolver->LaminarClosures(lamParam, u[0], u[1], bl->Ret[iPoint], bl->invBnd->GetMe(iPoint), bl->name);
        bl->Hstar[iPoint] = lamParam[0];
        bl->Hstar2[iPoint] = lamParam[1];
        bl->Hk[iPoint] = lamParam[2];
        bl->Cf[iPoint] = lamParam[3];
        bl->Cd[iPoint] = lamParam[4];
        bl->Delta[iPoint] = lamParam[5];
        bl->Cteq[iPoint] = lamParam[6];
        bl->Us[iPoint] = lamParam[7];
    }

    else if (bl->Regime[iPoint] == 1)
    {
        // Turbulent closure relations.
        std::vector<double> turbParam(8, 0.);
        bl->closSolver->TurbulentClosures(turbParam, u[0], u[1], u[4], bl->Ret[iPoint], bl->invBnd->GetMe(iPoint), bl->name);
        bl->Hstar[iPoint] = turbParam[0];
        bl->Hstar2[iPoint] = turbParam[1];
        bl->Hk[iPoint] = turbParam[2];
        bl->Cf[iPoint] = turbParam[3];
        bl->Cd[iPoint] = turbParam[4];
        bl->Delta[iPoint] = turbParam[5];
        bl->Cteq[iPoint] = turbParam[6];
        bl->Us[iPoint] = turbParam[7];
    }
    else
    {
        std::cout << "Wrong regime\n"
                  << std::endl;
        throw;
    }

    // Gradients computation.
    double dx = (bl->mesh->GetLoc(iPoint) - bl->mesh->GetLoc(iPoint - 1));
    double dxExt = (bl->mesh->GetExt(iPoint) - bl->mesh->GetExt(iPoint - 1));
    double dT_dx = (u[0] - bl->U[(iPoint - 1) * nVar + 0]) / dx;
    double dH_dx = (u[1] - bl->U[(iPoint - 1) * nVar + 1]) / dx;
    double due_dx = (u[3] - bl->U[(iPoint - 1) * nVar + 3]) / dx;
    double dCt_dx = (u[4] - bl->U[(iPoint - 1) * nVar + 4]) / dx;
    double dHstar_dx = (bl->Hstar[iPoint] - bl->Hstar[iPoint - 1]) / dx;

    double dueExt_dx = (bl->invBnd->GetVt(iPoint) - bl->invBnd->GetVt(iPoint - 1)) / dxExt;
    double ddeltaStarExt_dx = (bl->invBnd->GetDeltaStar(iPoint) - bl->invBnd->GetDeltaStar(iPoint - 1)) / dxExt;

    double c = 4 / (M_PI * dx);
    double cExt = 4 / (M_PI * dxExt);

    // Amplification routine.
    double dN_dx = 0.0;
    double Ax = 0.0;
    if (bl->xtrF == -1.0 && bl->Regime[iPoint] == 0)
    {
        // No forced transition and laminar regime.
        Ax = AmplificationRoutine(bl->Hk[iPoint], bl->Ret[iPoint], u[0]);
        dN_dx = (u[2] - bl->U[(iPoint - 1) * nVar + 2]) / dx;
    }

    double Mea = bl->invBnd->GetMe(iPoint);

    double Hstara = bl->Hstar[iPoint];
    double Hstar2a = bl->Hstar2[iPoint];
    double Hka = bl->Hk[iPoint];
    double Cfa = bl->Cf[iPoint];
    double Cda = bl->Cd[iPoint];
    double deltaa = bl->Delta[iPoint];
    double Cteqa = bl->Cteq[iPoint];
    double Usa = bl->Us[iPoint];

    // Space part.
    spaceVector(0) = dT_dx + (2 + u[1] - Mea * Mea) * (u[0] / u[3]) * due_dx - Cfa / 2;
    spaceVector(1) = u[0] * dHstar_dx + (2 * Hstar2a + Hstara * (1 - u[1])) * u[0] / u[3] * due_dx - 2 * Cda + Hstara * Cfa / 2;
    spaceVector(2) = dN_dx - Ax;
    spaceVector(3) = due_dx - c * (u[1] * dT_dx + u[0] * dH_dx) - dueExt_dx + cExt * ddeltaStarExt_dx;

    if (bl->Regime[iPoint] == 1)
        spaceVector(4) = (2 * deltaa / u[4]) * dCt_dx - 5.6 * (Cteqa - u[4] * dissipRatio) - 2 * deltaa * (4 / (3 * u[1] * u[0]) * (Cfa / 2 - (((Hka - 1) / (6.7 * Hka * dissipRatio)) * ((Hka - 1) / (6.7 * Hka * dissipRatio)))) - 1 / u[3] * due_dx);
    else
        spaceVector(4) = 0.0;

    // Time part.
    timeMatrix(0, 0) = u[1] / u[3];
    timeMatrix(0, 1) = u[0] / u[3];
    timeMatrix(0, 2) = 0.;
    timeMatrix(0, 3) = u[0] * u[1] / (u[3] * u[3]);
    timeMatrix(0, 4) = 0.;

    timeMatrix(1, 0) = (1 + u[1] * (1 - Hstara)) / u[3];
    timeMatrix(1, 1) = (1 - Hstara) * u[0] / u[3];
    timeMatrix(1, 2) = 0.;
    timeMatrix(1, 3) = (2 - Hstara * u[1]) * u[0] / (u[3] * u[3]);
    timeMatrix(1, 4) = 0.;

    timeMatrix(2, 0) = 0.;
    timeMatrix(2, 1) = 0.;
    timeMatrix(2, 2) = 1.;
    timeMatrix(2, 3) = 0.;
    timeMatrix(2, 4) = 0.;

    timeMatrix(3, 0) = -c * u[1];
    timeMatrix(3, 1) = -c * u[0];
    timeMatrix(3, 2) = 0.;
    timeMatrix(3, 3) = 1.;
    timeMatrix(3, 4) = 0.;

    timeMatrix(4, 0) = 0.;
    timeMatrix(4, 1) = 0.;
    timeMatrix(4, 2) = 0.;

    if (bl->Regime[iPoint] == 1)
    {
        timeMatrix(4, 3) = 2 * deltaa / (Usa * u[3] * u[3]);
        timeMatrix(4, 4) = 2 * deltaa / (u[3] * u[4] * Usa);
    }
    else
    {
        timeMatrix(4, 3) = 0.;
        timeMatrix(4, 4) = 1.0;
    }

    return timeMatrix.partialPivLu().solve(spaceVector);
}


double ViscFluxes::AmplificationRoutine(double Hk, double Ret, double theta) const
{
    double Dgr = 0.08;
    double Hk1 = 3.5;
    double Hk2 = 4.;
    double Hmi = (1/(Hk-1));
    double logRet_crit = 2.492*std::pow(1/(Hk-1.),0.43) + 0.7*(std::tanh(14*Hmi-9.24)+1.0); // value at which the amplification starts to grow
    
    double Ax = 0.;
    if (Ret <=0.)
        Ret = 1.;
    if (std::log10(Ret) < logRet_crit - Dgr)
        Ax = 0.;
    else
    {
        double Rnorm = (std::log10(Ret) - (logRet_crit-Dgr))/(2*Dgr);
        double Rfac = 0.;
        if (Rnorm <=0)
            Rfac = 0.;
        if (Rnorm >= 1)
            Rfac = 1.;
        else
            Rfac = 3*Rnorm*Rnorm - 2*Rnorm*Rnorm*Rnorm;

        // Normal envelope amplification rate
        double m = -0.05+2.7*Hmi-5.5*Hmi*Hmi+3*Hmi*Hmi*Hmi+0.1*std::exp(-20*Hmi);
        double arg = 3.87*Hmi-2.52;
        double dndRet = 0.028*(Hk-1)-0.0345*std::exp(-(arg*arg));
        Ax = (m*dndRet/theta)*Rfac;

        // Set correction for separated profiles
        if (Hk > Hk1)
        {
            double Hnorm = (Hk - Hk1)/(Hk2-Hk1);
            double Hfac = 0.;
            if (Hnorm >=1)
            {
                Hfac = 1.;
            }
            else
            {
                Hfac = 3*Hnorm*Hnorm - 2*Hnorm*Hnorm*Hnorm;
            }
            double Ax1 = Ax;
            double Gro = 0.3+0.35*std::exp(-0.15*(Hk-5));
            double Tnr = std::tanh(1.2*(std::log10(Ret)-Gro));
            double Ax2 = (0.086*Tnr - 0.25/std::pow((Hk-1),1.5))/theta;
            if (Ax2 < 0.)
            {
                Ax2 = 0.;
            }
            Ax = Hfac*Ax2 + (1-Hfac)*Ax1;
            if (Ax < 0.)
            {
                Ax = 0.;
            }
        }
    }
    return Ax;
}